/*
    GUT is a software that enables tunneling IP protocols over UDP.

    Copyright (C) 2010 by Aalto University


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "gut_vers.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <math.h>

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <linux/filter.h>
#include <sys/epoll.h>
#include <sys/time.h>

#include <linux/if_tun.h>
#include <linux/if.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <linux/tcp.h>
#include <linux/dccp.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <fcntl.h>
#include <unistd.h>

#include <time.h>
#include <signal.h>
#include <ifaddrs.h>

#ifndef IPPROTO_DCCP
#define IPPROTO_DCCP 33
#endif

#ifndef max
#define max( a, b ) ( ((a) > (b)) ? (a) : (b) )
#endif

#define DEBUG
#define GUT_P 4887
#define X_P 4888
#define BUFLEN 65500
#define MAXSTATES 100
#define GUTHDRLEN 4
#define OLD 10 /* in seconds when to remove a connection state */
#define CLEANUPTIMER 10
#define MAXIADDRS 128
#define MAX_TUN_ID 10
#define IPTABLES_PATH "/sbin/iptables"
#define TUN_PREFIX "tun"
#define IPTABLES_BASE_RULE_NUM 0x7F
#define MAX_IPTABLES_RULES 128
#define MARK_ID 3
#define TABLE_ID 203

#define IPT_OUTGOING_RULE_FMT "%s -t mangle -%c OUTPUT -j MARK --set-mark %d %s"

#define GUT_MIN_LEN(x) ((x->ihl * 4) + sizeof(struct udphdr) + GUTHDRLEN + 1)

struct gut_state
{
	unsigned int 		src;			/*source IP*/
	unsigned int		dst;			/*destination IP*/
	unsigned short 		protocol;		/*protocol number of native protocol*/
	unsigned short 		src_port;		/*source port of native packet*/
	unsigned short 		dst_port;		/*destination port of native packet*/
	unsigned short		gut_src_port;		/*source port of gut packet*/
	unsigned short 		gut_dst_port;		/*destination port of gut packet*/
	unsigned short 		sender;			/*0 = initiator, 1 = responder*/
	int 			sd;			/*file descriptor of initiator udp socket*/
	struct 			timeval last_active;	
};

struct gut_info
{
	int8_t 			gut_status; /* -1 = Error, 0 = Ok, 0x7F = Unknown encapsulation protocol */
	struct in_addr 		gut_src;
	struct in_addr 		gut_dst;
	uint16_t 		gut_src_port;
	uint16_t 		gut_dst_port;
	uint8_t	 		gut_encap_proto;
	uint16_t 		gut_encap_payload_len;
	uint16_t		gut_encap_src_port;
	uint16_t		gut_encap_dst_port;
	u_int8_t                ttl;
	u_int8_t                tos;
	u_int16_t               id;
	int                     sd;
	char                    *ip_opt;
	socklen_t               ip_opt_len;	
};

static struct 			gut_state states[MAXSTATES];
static struct in_addr 		iaddrs[MAXIADDRS];
static char 			str_src[INET6_ADDRSTRLEN],
				str_dst[INET6_ADDRSTRLEN],
				str_system[512];
static char 			* ipt_outgoing_rules[MAX_IPTABLES_RULES];
static int 			n_ipt_incoming_rules,
				n_ipt_outgoing_rules;

/* Configuration */
static char 			* iptables_path;
static char 			* tun_device_name;
static struct in_addr   	tun_device_addr;
static uint16_t 		gut_dst_port,
				gut_src_port,
				gut_mark_id,
				gut_table_id;

static void
iptables_rules(char cmd)
{
	int i;

	/* Set the OUTGOING chain rules */
	for (i = 0; i < n_ipt_outgoing_rules; i++)
	{
		printf("Set the OUTGOING chain rules\n");
		if (snprintf(str_system, sizeof(str_system), IPT_OUTGOING_RULE_FMT,
					iptables_path, cmd, gut_mark_id, ipt_outgoing_rules[i]) < sizeof(str_system))
		{
			 printf("%s\n", str_system);
			system(str_system);
		}
	}
		if (cmd == 'I')
		{
			/* Insert An ip routing rule for the traffic*/
			if (snprintf(str_system, sizeof(str_system), "ip rule add fwmark %d table %d",
						gut_mark_id, gut_table_id) < sizeof(str_system))
			{
				printf("%s\n", str_system);
			    system(str_system);
			}
		    /*Insert An ip forwarding table entry to the <table_id> table, redirecting all traffic in <table_id> to tun device*/
		    if (snprintf(str_system, sizeof(str_system), "ip route add table %d default dev %s",
						gut_table_id, tun_device_name) < sizeof(str_system))
			{
		    	printf("%s\n", str_system);
		    	system(str_system);
			}
		}
		else
		{
			/* Delete An ip routing rule for the traffic*/
			if (snprintf(str_system, sizeof(str_system), "ip rule del fwmark %d table %d",
						gut_mark_id, gut_table_id) < sizeof(str_system))
			{
				printf("%s\n", str_system);
				system(str_system);
			}
			/*Delete An ip forwarding table entry to the <table_id> table, redirecting all traffic in <table_id> to tun device*/
			if (snprintf(str_system, sizeof(str_system), "ip route del table %d default dev %s",
						gut_table_id, tun_device_name) < sizeof(str_system))
			{
				printf("%s\n", str_system);
		        system(str_system);
			}
		}

}

static void
signal_handler(int signal)
{
	switch (signal)
	{
		case SIGTERM:
		case SIGINT:
			iptables_rules('D');
			exit(1);
	}
}

static void
diep(char * s)
{
	perror(s);
	signal_handler(SIGTERM);
}

static void
disable_rp_filter(void)
{
	char zero;
	FILE * proc_fd;

	zero = '0';

	if (snprintf(str_system, sizeof(str_system), "/proc/sys/net/ipv4/conf/%s/rp_filter", tun_device_name) < sizeof(str_system))
	{
		if ((proc_fd = fopen(str_system, "w")) == NULL)
			diep("Unable to open proc filesystem handle");

		if (fwrite(&zero, sizeof(zero), 1, proc_fd) != 1)
			diep("Unable to disable Reverse Path filtering for TUN device");

		fclose(proc_fd);
	}
	else
		diep("Unable to disable Reverse Path Filtering for TUN device");

	if (snprintf(str_system, sizeof(str_system), "/proc/sys/net/ipv4/conf/all/rp_filter") < sizeof(str_system))
	{
		if ((proc_fd = fopen(str_system, "w")) == NULL)
			diep("Unable to open proc filesystem handle");

		if (fwrite(&zero, sizeof(zero), 1, proc_fd) != 1)
			diep("Unable to disable Reverse Path filtering for ALL");

		fclose(proc_fd);
	}
	else
		diep("Unable to disable Reverse Path Filtering for ALL");

}

static void
getiaddrs(void)
{
	struct sockaddr_in * sin;
	struct ifaddrs * ifaddrs,
				   * p_if;
	int i;

	getifaddrs(&ifaddrs);  /* Create a linked list of `struct ifaddrs' structures, one for each
				network interface on the host machine.  If successful, store the
				list in *IFAP and return 0.  On errors, return -1 and set `errno'.*/
	memset(iaddrs, 0, sizeof(struct in_addr) * MAXIADDRS);
	i = 0;

	for (p_if = ifaddrs; p_if && i < MAXIADDRS; p_if = p_if->ifa_next)
	{
		if (p_if->ifa_addr && p_if->ifa_netmask)
		{
			sin = (struct sockaddr_in *) p_if->ifa_addr;
			if (sin->sin_family != AF_INET ||
				IN_BADCLASS(ntohl(sin->sin_addr.s_addr)) ||
				sin->sin_addr.s_addr == 0 ||
				ntohl(sin->sin_addr.s_addr) == INADDR_LOOPBACK)
				continue;

			iaddrs[i++] = sin->sin_addr;   /*put ip address of each interface into array*/
		}

	}

	freeifaddrs(ifaddrs);
}

static int
localaddress(uint32_t addr)
{
	int i;

	for (i = 0; i < MAXIADDRS; i++)
	{
		if (iaddrs[i].s_addr == addr)
			return i;

		/* Break out when we run out of addresses to check */
		if (iaddrs[i].s_addr == 0)
			return -1;
	}

	return -1;
}

static uint16_t
in_cksum(uint16_t * buf, ssize_t len)
{
	int nleft;
	int sum;
	uint16_t * w;
	uint16_t answer;

	answer = 0; sum = 0;
	w = buf;
	nleft = len;

	while (nleft > 1)
	{
		sum += *buf++;
		nleft -= 2;
	}

	if (nleft == 1) {
		*((uint8_t *) &answer) = *((uint8_t *) buf);
		sum += answer;
	}

	sum = (sum >> 16) + (sum & 0xFFFF);
	sum += (sum >> 16);
	answer = ~sum;

	return (answer);
}


static uint16_t
in_cksum_udp(uint32_t src, uint32_t dst, uint16_t * udp_payload,
		uint16_t payload_len)
{
	struct psd_udp
	{
		uint32_t src;
		uint32_t dst;
		uint8_t pad;
		uint8_t proto;
		uint16_t udp_len;
	};

	uint8_t        	pseudo_udp_buf[payload_len + sizeof(struct psd_udp) +
		((payload_len % 2 > 0) ? 1 : 0)];
	struct psd_udp 	* p_pseudo_udp;

	p_pseudo_udp = (struct psd_udp *) pseudo_udp_buf;
	memset(pseudo_udp_buf, 0, sizeof(pseudo_udp_buf));

	p_pseudo_udp->src = src;
	p_pseudo_udp->dst = dst;
	p_pseudo_udp->pad = 0;
	p_pseudo_udp->proto = IPPROTO_UDP;
	p_pseudo_udp->udp_len = htons(payload_len);

	memcpy(p_pseudo_udp + 1, udp_payload, payload_len);

	return in_cksum((uint16_t *) pseudo_udp_buf, sizeof(pseudo_udp_buf));
}

static uint16_t
in_cksum_dccp(uint32_t src, uint32_t dst, uint16_t * dccp_payload,
		uint16_t payload_len)
{
	struct psd_dccp
	{
		uint32_t src;
		uint32_t dst;
		uint8_t pad;
		uint8_t proto;
		uint16_t dccp_len;
	};

	uint8_t        	* pseudo_dccp_buf;
	struct psd_dccp * p_pseudo_dccp;
	struct dccp_hdr * p_dccp;
	size_t 			dccp_buf_len;
	uint16_t 		sum;

	dccp_buf_len = sizeof(struct psd_dccp);

	p_dccp = (struct dccp_hdr *) dccp_payload;

	if (p_dccp->dccph_cscov > 0)
	{
		dccp_buf_len += p_dccp->dccph_doff * 4;
		dccp_buf_len += (p_dccp->dccph_cscov - 1) * 4;

		if (dccp_buf_len > payload_len)
			return 0;
	}
	else
		dccp_buf_len += payload_len + (((payload_len % 2) > 0) ? 1 : 0);

	pseudo_dccp_buf = malloc(dccp_buf_len);

	p_pseudo_dccp = (struct psd_dccp *) pseudo_dccp_buf;
	memset(pseudo_dccp_buf, 0, sizeof(pseudo_dccp_buf));

	p_pseudo_dccp->src = src;
	p_pseudo_dccp->dst = dst;
	p_pseudo_dccp->pad = 0;
	p_pseudo_dccp->proto = IPPROTO_DCCP;
	p_pseudo_dccp->dccp_len = htons(payload_len);

	memcpy(p_pseudo_dccp + 1, dccp_payload, dccp_buf_len - sizeof(struct psd_dccp));

	sum = in_cksum((uint16_t *) pseudo_dccp_buf, dccp_buf_len);
	free(pseudo_dccp_buf);
	return sum;
}

static uint16_t
in_cksum_tcp(uint32_t src, uint32_t dst, uint16_t * tcp_payload,
		uint16_t payload_len)
{
	struct psd_tcp
	{
		uint32_t src;
		uint32_t dst;
		uint8_t pad;
		uint8_t proto;
		uint16_t tcp_len;
	};

	uint8_t        	pseudo_tcp_buf[payload_len + sizeof(struct psd_tcp) +
		((payload_len % 2 > 0) ? 1 : 0)];
	struct psd_tcp 	* p_pseudo_tcp;

	p_pseudo_tcp = (struct psd_tcp *) pseudo_tcp_buf;
	memset(pseudo_tcp_buf, 0, sizeof(pseudo_tcp_buf));

	p_pseudo_tcp->src = src;
	p_pseudo_tcp->dst = dst;
	p_pseudo_tcp->pad = 0;
	p_pseudo_tcp->proto = IPPROTO_TCP;
	p_pseudo_tcp->tcp_len = htons(payload_len);

	memcpy(p_pseudo_tcp + 1, tcp_payload, payload_len);

	return in_cksum((uint16_t *) pseudo_tcp_buf, sizeof(pseudo_tcp_buf));
}

static struct gut_info
parse_native_gut_info(char * buf, size_t len)
{
	
	struct udphdr 		* p_udp;	/*UDP header*/
	struct iphdr 		* p_ip;		/*IP header*/
	struct dccp_hdr 	* p_dccp;	/*DCCP header*/	
	struct tcphdr		* p_tcp;	/*TCP header*/

	struct gut_info 	gut_info;	/*GUT information*/

	memset(&gut_info, 0, sizeof(gut_info));	/*clean up gut_info structure*/

	if (len < sizeof(struct iphdr))		/*if buffer length less than length of IP_header, error happens*/
	{
		gut_info.gut_status = -1;
		return gut_info;
	}

	p_ip = (struct iphdr *) buf;		/*cast buf into "struct iphdr" type*/

	/*tot_len: total length in IPv4 Header (bytes); ihl: Internet header length, telling the number of 32-bit words in the header
	 *p_ip->ihl *4: total length of IP header (bytes); 32-bit words = 4 bytes * 8-bits*/
	gut_info.gut_encap_payload_len = ntohs(p_ip->tot_len) - p_ip->ihl * 4;
	gut_info.gut_src.s_addr = p_ip->saddr;
	gut_info.gut_dst.s_addr = p_ip->daddr;
	gut_info.gut_encap_proto = p_ip->protocol;

	switch (gut_info.gut_encap_proto)
	{
		case IPPROTO_UDP:
			if (len < p_ip->ihl * 4 + sizeof(struct udphdr) + 1)
			{
				gut_info.gut_status = -1;
				return gut_info;
			}

			p_udp = (struct udphdr *) (buf + p_ip->ihl * 4);
			gut_info.gut_encap_src_port = ntohs(p_udp->source);
			gut_info.gut_encap_dst_port = ntohs(p_udp->dest);
			break;

		case IPPROTO_DCCP:
			p_dccp = (struct dccp_hdr *) (buf + p_ip->ihl * 4);

			if (len < p_ip->ihl * 4 + sizeof(struct dccp_hdr) ||			
				len < p_ip->ihl * 4 + sizeof(struct dccp_hdr) + dccp_packet_hdr_len(p_dccp->dccph_type))	/*dccph_type: Additional Fields (depending on type)*/
			{
				gut_info.gut_status = -1;
				return gut_info;
			}

			gut_info.gut_encap_src_port = ntohs(p_dccp->dccph_sport);
			gut_info.gut_encap_dst_port = ntohs(p_dccp->dccph_dport);
			break;

		case IPPROTO_ICMP:
			break;
			
		case IPPROTO_TCP:
			p_tcp = (struct tcphdr *) (buf + p_ip->ihl * 4);		/*get tcp header information*/
			if (len < p_ip->ihl * 4 + sizeof(struct tcphdr))		/*handling errors*/
			{
				gut_info.gut_status = -1;
				return gut_info;
			}

			gut_info.gut_encap_src_port = ntohs(p_tcp->source);	/*copy source and destination port of sctp packet*/
			gut_info.gut_encap_dst_port = ntohs(p_tcp->dest);
			break;

		default:
			gut_info.gut_status = 0x7F;
	}

	return gut_info;

}

static struct gut_info
parse_gut_info(struct msghdr* rcv_msg, size_t len)
{
	char 			* p_payload,
				* p_gut;
	struct udphdr 		* p_udp;
	struct dccp_hdr 	* p_dccp;
	struct tcphdr		* p_tcp;

	struct gut_info 	gut_info;

	struct in_pktinfo 	*packet_info;
	struct cmsghdr 		*cmsg;
	struct sockaddr_in 	*rcv_msg_sinfo;

	
	rcv_msg_sinfo = (struct sockaddr_in *) (rcv_msg->msg_name);
	memset(&gut_info, 0, sizeof(gut_info));

	gut_info.ip_opt = NULL;
		
	for( cmsg=CMSG_FIRSTHDR(rcv_msg); cmsg != NULL; cmsg=CMSG_NXTHDR(rcv_msg,cmsg) )
		{	
			
			if (cmsg->cmsg_type == IP_PKTINFO)
			{
				
				packet_info = (struct in_pktinfo*) CMSG_DATA(cmsg);
				
			}
			else if (cmsg->cmsg_type == IP_TOS) 
			{
				
				gut_info.tos = *((u_int8_t *) CMSG_DATA(cmsg));
				
			}
			else if (cmsg->cmsg_type == IP_TTL) 
			{
				gut_info.ttl = *((u_int8_t *) CMSG_DATA(cmsg));
				
			}
			else if (cmsg->cmsg_type == IP_OPTIONS) 
			{
				gut_info.ip_opt = (char *) CMSG_DATA(cmsg);
				gut_info.ip_opt_len = cmsg->cmsg_len-CMSG_ALIGN(sizeof(struct cmsghdr));
				printf("Got IP_RECVOPTS!\n");
			}
		}	
	
	
	p_gut = (char *) rcv_msg->msg_iov[0].iov_base;
	p_payload = p_gut + GUTHDRLEN;

	gut_info.gut_encap_payload_len = len - GUTHDRLEN;
	gut_info.gut_src.s_addr = rcv_msg_sinfo->sin_addr.s_addr;
	gut_info.gut_src_port = ntohs(rcv_msg_sinfo->sin_port);
	gut_info.gut_dst.s_addr = packet_info->ipi_addr.s_addr;
	gut_info.gut_encap_proto = *(p_payload - 1);

	switch (gut_info.gut_encap_proto)
	{
		case IPPROTO_UDP:
			p_udp = (struct udphdr *) p_payload;
			gut_info.gut_encap_src_port = ntohs(p_udp->source);
			gut_info.gut_encap_dst_port = ntohs(p_udp->dest);
			break;

		case IPPROTO_DCCP:
			p_dccp = (struct dccp_hdr *) p_payload;
			gut_info.gut_encap_src_port = ntohs(p_dccp->dccph_sport);
			gut_info.gut_encap_dst_port = ntohs(p_dccp->dccph_dport);

			break;

		case IPPROTO_ICMP:
			break;
						
		case IPPROTO_TCP:
			p_tcp = (struct tcphdr *) p_payload;
			gut_info.gut_encap_src_port = ntohs(p_tcp->source);
			gut_info.gut_encap_dst_port = ntohs(p_tcp->dest);
			break;

		default:
			gut_info.gut_status = 0x7F;
	}

	return gut_info;
}

static int
encapsulate_datagram(char * buf, size_t buf_len, const struct gut_info * gi)
{
	struct iphdr 	* p_ip;
	char 		* p_gut;
	char 		* p_payload,
			* p_encap_payload;

	p_ip = (struct iphdr *) buf;

	if (gi->gut_encap_payload_len + sizeof(struct udphdr) + GUTHDRLEN > buf_len)
		return -1;
	p_payload = (char *) (buf + p_ip->ihl * 4);
	p_gut = buf;
	p_encap_payload = (p_gut + GUTHDRLEN);

	/* After parsing original transport information out, encapsulate. */
	memmove(p_encap_payload,
		p_payload,
		gi->gut_encap_payload_len);

	*(p_encap_payload - 1) = gi->gut_encap_proto;

	return (gi->gut_encap_payload_len + GUTHDRLEN);
}

static int
decapsulate_datagram(struct msghdr *rcv_msg, const struct gut_info * gi)
{
	char 			* p_payload,
				* p_gut;
	struct udphdr 		* p_udp;
	struct iphdr 		* p_ip;
	struct dccp_hdr 	* p_dccp;
	struct tcphdr		* p_tcp;
	int                      ihl = 5;
	uint16_t 		csum;
	uint8_t 		protocol;
	p_gut = (char *) rcv_msg->msg_iov[0].iov_base;
	p_payload = p_gut + GUTHDRLEN;
	protocol = *(p_payload - 1);
        p_ip = (struct iphdr *)rcv_msg->msg_iov[0].iov_base;
	if (gi->ip_opt != NULL)
		ihl += gi->ip_opt_len;

	/* Remove the GUT header from the packet */
	memmove(p_gut + ihl*4,
			p_payload,
			gi->gut_encap_payload_len);

	switch (protocol)
	{
		case IPPROTO_UDP:
			p_udp = (struct udphdr *) (p_gut + ihl*4 );

			/* Recalculate UDP checksum for decapsulated packet */
			p_udp->check = 0;
			csum = in_cksum_udp(gi->gut_src.s_addr, gi->gut_dst.s_addr,
					(uint16_t *) p_udp, ntohs(p_udp->len));
			p_udp->check = (csum > 0) ? csum : 0xFFFF;
			break;

		case IPPROTO_DCCP:
			p_dccp = (struct dccp_hdr *) (p_gut + ihl*4);

			p_dccp->dccph_checksum = 0;
			csum = in_cksum_dccp(gi->gut_src.s_addr, gi->gut_dst.s_addr,
					(uint16_t *) p_dccp, gi->gut_encap_payload_len);

			p_dccp->dccph_checksum = csum;
			break;

		case IPPROTO_ICMP:
			break;

		case IPPROTO_TCP:
			p_tcp = (struct tcphdr *) (p_gut + ihl*4);

			p_tcp->check = 0;			/*typedef __u16 __bitwise __sum16;*/
			csum = in_cksum_tcp(gi->gut_src.s_addr,gi->gut_dst.s_addr,
					(uint16_t *) p_tcp, gi->gut_encap_payload_len);

			p_tcp->check = csum;
			break;

		default:
			break;
	}
	
	/* Create IP Header*/
	p_ip->version = 4;
	p_ip->ihl = ihl;
	p_ip->check = 0;
	p_ip->protocol = protocol;
	p_ip->tot_len = htons(gi->gut_encap_payload_len + p_ip->ihl * 4);
	p_ip->saddr = gi->gut_src.s_addr;
	p_ip->daddr = gi->gut_dst.s_addr;
	p_ip->id = 60052;
	p_ip->frag_off = 64;
	p_ip->tos = gi->tos;
	p_ip->ttl = gi->ttl;	
	p_ip->check = in_cksum((uint16_t *) p_ip, p_ip->ihl * 4);
	if(gi->ip_opt != NULL)
		memcpy(p_gut + 20, gi->ip_opt, gi->ip_opt_len);
	return ntohs(p_ip->tot_len);
}

static int
find_flow(unsigned long addr, unsigned char protocol, unsigned short src_port,
		unsigned short dst_port)
{
	int i;
	for(i=0;i<MAXSTATES;i++)
	{
		if((addr == states[i].src) &&
				(protocol == states[i].protocol) &&
				(src_port == states[i].src_port) &&
				(dst_port == states[i].dst_port) )
			return (i);
	}
	return -1;
}

static int
store_info(unsigned long addr, unsigned char protocol, unsigned short src_port,
		unsigned short dst_port, unsigned short sender, unsigned short gut_tx_port, int sd)
{
	int i;
	for(i=0;i<MAXSTATES;i++)
	{
		if(states[i].src == 0)
		{
			states[i].src = addr;
			states[i].protocol = protocol;
			states[i].src_port = src_port;
			states[i].dst_port = dst_port;
			states[i].sender = sender;
			states[i].gut_dst_port = gut_tx_port;
			states[i].sd = sd;
			gettimeofday(&states[i].last_active,NULL);
			return 1;
		}
	}
	return 0;
}

static void
clear_state(int j)
{
	states[j].src = 0;
	states[j].protocol = 0;
	states[j].src_port = 0;
	states[j].dst_port = 0;
	states[j].sender = 0;
	states[j].last_active.tv_sec=0;
	states[j].last_active.tv_usec=0;
	states[j].sd = -1;
}
/*TUN (as in network TUNnel) simulates a network layer device
 * and it operates with layer 3 packets such as IP packets*/
static int
tun_alloc(uint32_t addr)
{
	struct ifreq ifr;
	struct sockaddr_in * sin;
	int i, fd, err, tmp_fd;

	if( (fd = open("/dev/net/tun", O_RDWR)) < 0 )
		diep("Unable to open tun device");

	memset(&ifr, 0, sizeof(ifr));

	/* Flags: IFF_TUN   - TUN device (no Ethernet headers)
	 *        IFF_TAP   - TAP device
	 *
	 *        IFF_NO_PI - Do not provide packet information
	 */

	ifr.ifr_flags = IFF_TUN;
	ifr.ifr_flags |= IFF_NO_PI;

	for (i = 0; i < MAX_TUN_ID; i++)
	{
		snprintf(ifr.ifr_name, IFNAMSIZ, "%s%d", tun_device_name, i);		/*tun device [0,1,...,MAX_TUN_ID]*/
		/*Perform the I/O control operation specified by REQUEST on FD*/
		if ((err = ioctl(fd, TUNSETIFF, (void *) &ifr)) < 0)
			continue;
		else
			break;
	}

	if (i == MAX_TUN_ID)
	{
		close(fd);
		diep("Unable to open a valid TUN device");
	}
	else
	{
		tun_device_name = realloc(tun_device_name, strlen(ifr.ifr_name) + 1);
		strcpy(tun_device_name, ifr.ifr_name);
		tun_device_name[strlen(ifr.ifr_name)] = '\0';
	}

	/* Set the interface address then. */
	ifr.ifr_flags = 0;

	sin = (struct sockaddr_in *) &ifr.ifr_addr;   /*pointer to ifr.ifr_addr*/

	sin->sin_family = AF_INET;
	sin->sin_addr.s_addr = addr;

	tmp_fd = socket(AF_INET, SOCK_DGRAM, 0);

	if (ioctl(tmp_fd, SIOCSIFADDR, (void *) &ifr) < 0)
		diep("SIOCSIFADDR");

	if (ioctl(tmp_fd, SIOCGIFINDEX, (void *) &ifr) < 0)
		diep("SIOCGIFINDEX");

	if (ioctl(tmp_fd, SIOCGIFFLAGS, (void *) &ifr) < 0)
		diep("SIOCGIFFLAGS");

	ifr.ifr_flags |= IFF_RUNNING;
	ifr.ifr_flags |= IFF_UP;

	if (ioctl(tmp_fd, SIOCSIFFLAGS, (void *) &ifr) < 0)
		diep("SIOCSIFFLAGS");

	close(tmp_fd);

	return fd;
}

static void
usage(char * p)
{
	fprintf(stderr, "%s\n", gut_vers);	
	fprintf(stderr, "Usage: %s -a tun_addr [-ihtsdnov]\n", p);
}

static void
help(char * p)
{
	usage(p);

	fprintf(stderr, "Options:\n");
	fprintf(stderr, "  %-13.13s Set TUN Interface IPv4 Address\n", "-a ADDR");
	fprintf(stderr, "  %-13.13s Set TUN Interface prefix name (default %s)\n", "-i PREFIX", TUN_PREFIX);
	fprintf(stderr, "  %-13.13s Set full path to iptables executable (default %s)\n", "-t PATH", IPTABLES_PATH);
	fprintf(stderr, "  %-13.13s Set GUT_P (default %d)\n", "-d PORT", GUT_P);
	fprintf(stderr, "  %-13.13s Add iptables OUTGOING rule match options\n", "-o STRING");
	fprintf(stderr, "  %-13.13s Set iptables OUTGOING rule mark id (default %d)\n", "-m MARK_ID", MARK_ID);
	fprintf(stderr, "  %-13.13s Set iptables OUTGOING rule routing table id (default %d)\n", "-r TABLE_ID", TABLE_ID);

}


int
main(int argc, char *argv[])
{
	int 			sm, gut_sd, gut_rd, tun_fd, n, i, j,p, payload_len, epfd,nfds;
	socklen_t 		slen;
	struct timeval 		now;
	struct sockaddr_in 	si_dst, si_gut;
	struct gut_info 	gut_info;		/*information of gut packet, structure type*/

	char 			cbuf[BUFLEN], buf[BUFLEN];
	struct iphdr 		* ip_hdr;
	struct sigaction 	act, old_act;
	struct msghdr rcv_msg;
	struct iovec iov;
	struct sockaddr_in rcv_msg_sinfo;
	struct in_pktinfo *packet_info;
	struct epoll_event ev;
	struct epoll_event events[1000];
	struct in_addr src;
	sm = 1;
	slen = sizeof(struct sockaddr_in);
	
	/* Set some defaults */
	tun_device_name = malloc(strlen(TUN_PREFIX) + 1); /*device name*/
	strcpy(tun_device_name, TUN_PREFIX);

	iptables_path = malloc(strlen(IPTABLES_PATH + 1)); /*path of iptable*/
	strcpy(iptables_path, IPTABLES_PATH);

	gut_src_port = X_P;
	gut_dst_port = GUT_P;
	n_ipt_incoming_rules = 0;
	n_ipt_outgoing_rules = 0;
	gut_mark_id = MARK_ID;
	gut_table_id = TABLE_ID;

	getiaddrs(); 				/*get ip address of each network interface; iaadrs : struct in_addr[]*/
	srand(time(NULL));			/* Seed the random number generator with the given number; using time as seed to diff generators */

	for(i=0;i<MAXSTATES;i++)
		clear_state(i);			/*clear all states; struct gut_state[]*/

	if (argc < 2)
	{
		usage(argv[0]);
		return 1;;
	}

	while ((i = getopt(argc, argv, "hi:a:t:s:d:o:n:m:r:v")) != -1)
	{
		switch (i)
		{
			case 'a':
				if (inet_pton(AF_INET, optarg, &tun_device_addr) <= 0)		/* set ip address of tun device*/
					diep("Invalid TUN Device IPv4 address.");
				break;
			case 'i':
				if ((n = strlen(optarg)) < 1)
					diep("Invalid TUN Device prefix name");

				tun_device_name = realloc(tun_device_name, n + 1);
				strcpy(tun_device_name, optarg);
				tun_device_name[n] = '\0';
				break;
			case 't':
				if ((n = strlen(optarg)) < 1)
					diep("Invalid iptables path");

				iptables_path = realloc(iptables_path, n + 1);
				strcpy(iptables_path, optarg);
				iptables_path[n] = '\0';
				break;
	
			case 'd':
				if ((n = strlen(optarg)) < 1)				/*Set GUT UDP destination port*/
					diep("Invalid GUT UDP Destination port");

				p = atoi(optarg);
				if (p < 0 || p > 65535)
					diep("Invalid GUT UDP Destination port");

				gut_dst_port = p;
				break;

			case 'o':
				if ((n = strlen(optarg)) < 1)
					diep("Invalid OUTGOING iptables match option");

				if (n_ipt_outgoing_rules < MAX_IPTABLES_RULES)
				{
					ipt_outgoing_rules[n_ipt_outgoing_rules] = malloc(n + 1);
					strcpy(ipt_outgoing_rules[n_ipt_outgoing_rules], optarg);
					ipt_outgoing_rules[strlen(optarg)] = '\0';
					n_ipt_outgoing_rules++;
				}
				else
				{
					fprintf(stderr, "Maximum amount of OUTGOING iptables rules reached\n");
					signal_handler(SIGTERM);
				}

				break;			

			case 'm':
			if ((n = strlen(optarg)) < 1)
				diep("Invalid OUTGOING iptables mark id option");

			p = atoi(optarg);
			gut_mark_id =  p;
			    break;


			case 'r':
			if ((n = strlen(optarg)) < 1)
				diep("Invalid OUTGOING routing table id option");

			p = atoi(optarg);
			gut_table_id =  p;
			    break;

			case 'h':
			case '?':
				help(argv[0]);
				return 1;
			case 'v':
				fprintf(stderr,"%s\n",gut_vers);
				return 1;
			default:
				diep("Unknown option");
		}
	}

	/* Signal handler stuff */
	memset(&act, 0, sizeof(act));
	act.sa_handler = signal_handler;

	sigfillset(&act.sa_mask);
	sigdelset(&act.sa_mask, SIGINT);

	if (sigaction(SIGINT, &act, &old_act) != 0)
		diep("Unable to set signal handler");

	sigfillset(&act.sa_mask);
	sigdelset(&act.sa_mask, SIGTERM);

	if (sigaction(SIGTERM, &act, &old_act) != 0)
		diep("Unable to set signal handler");

	

	/*
	 * Assign ip address and name to tun device
	 * return fd of tun device file
	 */
	tun_fd = tun_alloc(tun_device_addr.s_addr);

	disable_rp_filter();			/*Disable reserver path filter*/

       	printf("Set the TUN Device MTU to 65500\n");
	if (snprintf(str_system, sizeof(str_system), "ip link set dev %s mtu 65500",
			tun_device_name) < sizeof(str_system))
	{
		printf("%s\n", str_system);
		system(str_system);
	}
		
	if ( (gut_rd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
		diep("GUT Server Socket creation error");
	
	bzero(&si_gut, sizeof(si_gut));
	si_gut.sin_family = AF_INET;
	si_gut.sin_port = htons(gut_dst_port);
	si_gut.sin_addr.s_addr = htonl(INADDR_ANY);
	if (bind(gut_rd, (struct sockaddr *) &si_gut, sizeof(si_gut))==-1)
		diep("bind");
	
	
	if ( setsockopt(gut_rd, IPPROTO_IP, IP_PKTINFO, &sm, sizeof(sm)) == -1 )
		diep("Unable to set IP_PKTNFO on GUT Receive Socket");
	
	if ( setsockopt(gut_rd, IPPROTO_IP, IP_RECVTOS, &sm, sizeof(sm)) == -1 )
		diep("Unable to set IP_RECVTOS on GUT Receive Socket");
	
	if ( setsockopt(gut_rd, IPPROTO_IP, IP_RECVTTL, &sm, sizeof(sm)) == -1 )
		diep("Unable to set IP_RECVTTL on GUT Receive Socket");
	
	if ( setsockopt(gut_rd, IPPROTO_IP, IP_RECVOPTS, &sm, sizeof(sm)) == -1 )
		diep("Unable to set IP_RECVOPTS on GUT Receive Socket");

	if ((epfd = epoll_create(1000)) < 0)
		diep("epoll_create");
	ev.events = EPOLLIN;
        ev.data.fd = tun_fd;
	if (epoll_ctl(epfd, EPOLL_CTL_ADD, tun_fd, &ev) < 0) 
                diep("epoll_ctl");
	ev.events = EPOLLIN;
        ev.data.fd = gut_rd;
	if (epoll_ctl(epfd, EPOLL_CTL_ADD, gut_rd, &ev) < 0)
                diep("epoll_ctl");

	memset((char *) &rcv_msg, 0, sizeof(struct msghdr));
	packet_info = NULL;
	rcv_msg.msg_control = cbuf;
	rcv_msg.msg_controllen = sizeof(cbuf);
	iov.iov_base = buf;
	iov.iov_len = sizeof(buf);
	rcv_msg.msg_iov = &iov;
	rcv_msg.msg_iovlen = 1;
	rcv_msg.msg_name = (void*) &rcv_msg_sinfo;
	rcv_msg.msg_namelen = sizeof(rcv_msg_sinfo);

	/*
	 * configure iptable with parameter -I
	 * iptable: linux kernal firewall
	*/
	iptables_rules('I');

	while(1)
	{
		now.tv_sec = CLEANUPTIMER;
		now.tv_usec = 0;

		nfds = epoll_wait(epfd, events, 1000, CLEANUPTIMER*1000);
		if(nfds < 0) diep("epoll_wait()");

    		for(j = 0; j < nfds; j++) 
		{
        		if(events[j].data.fd == tun_fd) 
			{
				if ((n = read(tun_fd, buf, sizeof(buf))) < 1)
					diep("read(tun)");

				if (n == sizeof(buf))
					diep("read() Payload too long");

				/* Check where the packet originated from. If it originated on
				* local host we need to encapsulate & send. If it originated from
			 	* an internet host we need to decapsulate & process locally.
			 	*/

				ip_hdr = (struct iphdr *) buf;

				if (localaddress(ip_hdr->saddr) > -1)					/*Check ip address of packet from where*/
				{
					#ifdef DEBUG
					printf("Incoming locally originated packet\n");
					#endif
					gut_info = parse_native_gut_info(buf, n);			/*Parse gut information from buffer; TODO: SCTP, TCP extension*/
					if (gut_info.gut_status == -1)
						continue;

					if ((i = find_flow(gut_info.gut_src.s_addr,
								gut_info.gut_encap_proto,
								gut_info.gut_encap_src_port,
								gut_info.gut_encap_dst_port)) > -1)
					{	
						#ifdef DEBUG
						printf("Existing flow, initiator\n");
						#endif
						gut_info.gut_dst_port = states[i].gut_dst_port;
						gut_info.gut_src_port = gut_info.gut_encap_src_port;
						gettimeofday(&states[i].last_active,NULL);
						gut_sd = states[i].sd;
					}
					else if ((i = find_flow(gut_info.gut_dst.s_addr,
								gut_info.gut_encap_proto,
								gut_info.gut_encap_dst_port,
								gut_info.gut_encap_src_port)) > -1)
					{
						#ifdef DEBUG
						printf("Existing flow, responder\n");
						#endif
						gut_info.gut_dst_port = states[i].gut_dst_port;
						gut_info.gut_src_port = gut_dst_port;
						gut_sd = gut_rd;
					}
					else
					{	
						printf("New flow for initiator\n");
						if ((gut_sd = socket(AF_INET, SOCK_DGRAM, 0)) == -1)		//Create GUT Socket
							diep("GUT Socket creation error");

						if ( setsockopt(gut_sd, IPPROTO_IP, IP_PKTINFO, &sm, sizeof(sm)) == -1 )
							diep("Unable to set IP_PKTNFO on GUT Receive Socket");
	
						if ( setsockopt(gut_sd, IPPROTO_IP, IP_RECVTOS, &sm, sizeof(sm)) == -1 )
							diep("Unable to set IP_RECVTOS on GUT Receive Socket");
	
						if ( setsockopt(gut_sd, IPPROTO_IP, IP_RECVTTL, &sm, sizeof(sm)) == -1 )
							diep("Unable to set IP_RECVTTL on GUT Receive Socket");
	
						if ( setsockopt(gut_sd, IPPROTO_IP, IP_RECVOPTS, &sm, sizeof(sm)) == -1 )
							diep("Unable to set IP_RECVOPTS on GUT Receive Socket");
						
						if(store_info(gut_info.gut_src.s_addr,
								gut_info.gut_encap_proto,
								gut_info.gut_encap_src_port,
								gut_info.gut_encap_dst_port, 1, 
								gut_dst_port, 
								gut_sd))
							printf("Stored info about this new flow\n");
						else
							diep("no more states for flows\n");

						gut_info.gut_dst_port = gut_dst_port;
						
					
						bzero(&si_gut, sizeof(si_gut));
						si_gut.sin_port = htons(gut_info.gut_encap_src_port);
						si_gut.sin_family = AF_INET;
						si_gut.sin_addr.s_addr = htonl(INADDR_ANY);
						
						if (bind(gut_sd, (struct sockaddr *) &si_gut, sizeof(si_gut))==-1)
							printf("bind error, let system choose source port.\n");
						ev.events = EPOLLIN;
        					ev.data.fd = gut_sd;						
						if (epoll_ctl(epfd, EPOLL_CTL_ADD, gut_sd, &ev) < 0)
                					diep("epoll set insertion error");			
					}

					payload_len = encapsulate_datagram(buf, sizeof(buf), &gut_info);
					bzero(&si_dst, sizeof(si_dst));
					si_dst.sin_family = AF_INET;
					si_dst.sin_port = htons(gut_info.gut_dst_port);
					si_dst.sin_addr = gut_info.gut_dst;
					if (sendto(gut_sd, buf,
							payload_len,
							0, (struct sockaddr *) &si_dst,
							sizeof(struct sockaddr_in)) == -1)
						diep("sendto()");

					getsockname(gut_sd, (struct sockaddr*)&si_gut, &slen);
					gut_info.gut_src_port = ntohs(si_gut.sin_port);
					#ifdef DEBUG
					printf("Local packet: %s:%d -> %s:%d, proto=%d len=%d\n",
						inet_ntop(AF_INET, &gut_info.gut_src, str_src,
							INET6_ADDRSTRLEN), gut_info.gut_encap_src_port,
						inet_ntop(AF_INET, &gut_info.gut_dst, str_dst,
							INET6_ADDRSTRLEN), gut_info.gut_encap_dst_port,
						gut_info.gut_encap_proto,
						gut_info.gut_encap_payload_len);

					printf("GUT   packet: %s:%d -> %s:%d, len=%d\n",
						inet_ntop(AF_INET, &gut_info.gut_src, str_src,
							INET6_ADDRSTRLEN), gut_info.gut_src_port,
						inet_ntop(AF_INET, &gut_info.gut_dst, str_dst,
							INET6_ADDRSTRLEN), gut_info.gut_dst_port,
						payload_len);
					#endif

				
					
				}
			}
			
			else
			{	
				#ifdef DEBUG
				printf("Incoming remote originated packet\n");	
				#endif
				if ( (n = recvmsg(events[j].data.fd, &rcv_msg, 0)) < 1)
					diep("recvmsg(events[j].data.fd)");
				gut_info = parse_gut_info(&rcv_msg, n);

				if (gut_info.gut_status == -1)
					continue;
				getsockname(events[j].data.fd, (struct sockaddr*)&si_gut, &slen);
				gut_info.gut_dst_port = ntohs(si_gut.sin_port);
				
				payload_len = decapsulate_datagram(&rcv_msg, &gut_info);

				/* Existing flow: @Responder */
				if ((i = find_flow(gut_info.gut_src.s_addr,
					gut_info.gut_encap_proto,
					gut_info.gut_encap_src_port,
					gut_info.gut_encap_dst_port)) > -1)
				{
					#ifdef DEBUG
					printf("Existing flow, responder\n");
					#endif
					gettimeofday(&states[i].last_active,NULL);
					states[i].gut_dst_port = gut_info.gut_src_port;
				}
				/* Existing flow: @Initiator */
				else if ((i = find_flow(gut_info.gut_dst.s_addr,
						gut_info.gut_encap_proto,
						gut_info.gut_encap_dst_port,
						gut_info.gut_encap_src_port)) > -1)
				{
					#ifdef DEBUG
					printf("Existing flow, initiator\n");
					#endif
					gettimeofday(&states[i].last_active,NULL);
					states[i].gut_dst_port = gut_info.gut_src_port;
				}
				/* New flow */
				else
				{
					printf("New flow for reponder\n");
					if(store_info(gut_info.gut_src.s_addr,
							gut_info.gut_encap_proto,
							gut_info.gut_encap_src_port,
							gut_info.gut_encap_dst_port, 0,
							gut_info.gut_src_port,
							-1))
						printf("Stored info about this new flow\n");
					else
						diep("no more states for flows\n");
					printf("Set the new OUTGOING chain rules\n");
					if (gut_info.gut_encap_proto == 33) 
					{
						if (snprintf(str_system, 
							     sizeof(str_system),
							     "%s -t mangle -I OUTPUT -j MARK --set-mark %d -p %d -m dccp --dport %d -d %s",
							     iptables_path, gut_mark_id, 
						             gut_info.gut_encap_proto, 
							     gut_info.gut_encap_src_port,
							     inet_ntop(AF_INET, &gut_info.gut_src, str_src, INET6_ADDRSTRLEN)) < sizeof(str_system))
						{
				 			printf("%s\n", str_system);
							system(str_system);
						}
					}
					else if (gut_info.gut_encap_proto == 1) 
					{
						if (snprintf(str_system, sizeof(str_system),
						    	     "%s -t mangle -I OUTPUT -j MARK --set-mark %d -p %d --icmp-type 0 -d %s",
						    	     iptables_path, gut_mark_id, gut_info.gut_encap_proto,
						             inet_ntop(AF_INET, &gut_info.gut_src, str_src, INET6_ADDRSTRLEN)) < sizeof(str_system))
						{
				 			printf("%s\n", str_system);
							system(str_system);
						}
					}
					else 
					{
						if (snprintf(str_system, 
							     sizeof(str_system),
							     "%s -t mangle -I OUTPUT -j MARK --set-mark %d -p %d --dport %d -d %s",
							     iptables_path, gut_mark_id, 
						             gut_info.gut_encap_proto, 
							     gut_info.gut_encap_src_port,
							     inet_ntop(AF_INET, &gut_info.gut_src, str_src, INET6_ADDRSTRLEN)) < sizeof(str_system))
						{
				 			printf("%s\n", str_system);
							system(str_system);
						}
					}
				
				}

				#ifdef DEBUG
				printf("GUT   packet %s:%d -> %s:%d, len=%d\n", inet_ntop(AF_INET,
						&gut_info.gut_src, str_src, INET6_ADDRSTRLEN),
						gut_info.gut_src_port, inet_ntop(AF_INET,
						&gut_info.gut_dst, str_dst, INET6_ADDRSTRLEN),
						gut_info.gut_dst_port, n);

				printf("Local packet %s:%d -> %s:%d, proto=%d len=%d\n",
						inet_ntop(AF_INET, &gut_info.gut_src, str_src,
						INET6_ADDRSTRLEN), gut_info.gut_encap_src_port,
						inet_ntop(AF_INET, &gut_info.gut_dst, str_dst,
						INET6_ADDRSTRLEN), gut_info.gut_encap_dst_port,
						gut_info.gut_encap_proto, payload_len);
				#endif

				/* Packet decapsulated ok, write it to tun device and hope for the* best. */
				if ((n = write(tun_fd, (char*)rcv_msg.msg_iov[0].iov_base, payload_len)) == -1)
					perror("write(tun_fd):");
				#ifdef DEBUG
				else
					printf("write(tun_fd): %d\n", n);
				#endif
			}
		}

		
		/* Go through the list of sessions and delete old ones */

		gettimeofday(&now,NULL);

		for (i=0;i<MAXSTATES;i++)
		{
			if(states[i].src > 0)
				if(now.tv_sec - states[i].last_active.tv_sec > OLD)
				{
					printf("Session %d has expired\n",i);
					
                         		if(states[i].sender == 1) 
					{
						epoll_ctl(epfd, EPOLL_CTL_DEL, states[i].sd, &ev);
						close(states[i].sd);
					}
					else 
					{
						src.s_addr = states[i].src;
						if (states[i].protocol == 33) 
						{    
							if (snprintf(str_system, sizeof(str_system),
							             "%s -t mangle -D OUTPUT -j MARK --set-mark %d -p %d -m dccp --dport %d -d %s",
								     iptables_path, gut_mark_id, states[i].protocol, states[i].src_port, 
								     inet_ntop(AF_INET, &src, str_src,INET6_ADDRSTRLEN)) < sizeof(str_system))
							{
			 					printf("%s\n", str_system);
								system(str_system);
							}
						}
						else if (states[i].protocol == 1)
						{
							if (snprintf(str_system, sizeof(str_system),
							             "%s -t mangle -D OUTPUT -j MARK --set-mark %d -p %d --icmp-type 0 -d %s",
								      iptables_path, gut_mark_id, states[i].protocol, 
								     inet_ntop(AF_INET, &src, str_src,INET6_ADDRSTRLEN)) < sizeof(str_system))
							{
			 					printf("%s\n", str_system);
								system(str_system);
							}
						}
						else 
						{
							if (snprintf(str_system, sizeof(str_system),
								     "%s -t mangle -D OUTPUT -j MARK --set-mark %d -p %d --dport %d -d %s",
							             iptables_path, gut_mark_id, states[i].protocol, states[i].src_port, 
							             inet_ntop(AF_INET, &src, str_src,INET6_ADDRSTRLEN)) < sizeof(str_system))
							{
			 					printf("%s\n", str_system);
								system(str_system);
							}
						}
					}
					
					clear_state(i);
				}
		}

	}

	return 0;
}
